import pandas as pd
import numpy as np

import pyranges as pr

import MySQLdb

from itertools import chain


def _exons(df):

    _df = df.drop(["Start", "End"], axis=1)

    cols = _df.columns.difference(['XS', 'XE'])
    exon_starts = _df['XS'].str.replace(",$", "").str.split(',')
    exon_ends = _df['XE'].str.replace(",$", "").str.split(',')

    exon_numbers = exon_starts.apply(lambda x:
                                     [i for i in range(1,
                                                       len(x) + 1)])

    ls = exon_starts.str.len()
    exon_starts = [int(i) for i in chain.from_iterable(exon_starts.tolist())]
    exon_ends = [int(i) for i in chain.from_iterable(exon_ends.tolist())]
    exon_starts = pd.Series(exon_starts, dtype=np.int32)
    exon_ends = pd.Series(exon_ends, dtype=np.int32)
    exon_numbers = pd.Series(
        list(chain.from_iterable(exon_numbers.tolist())), dtype="category")
    exons = (_df.loc[_df.index.repeat(ls), cols].assign(
        Start=exon_starts, End=exon_ends,
        Feature="exon").reset_index(drop=True).assign(ExonNumber=exon_numbers))

    return exons


def ucsc(genome, query):

    host = "genome-mysql.cse.ucsc.edu"
    user = "genome"
    db = genome

    conn = MySQLdb.Connection(host=host, user=user, db=db)

    df = pd.read_sql(query, conn)

    conn.close()

    return df


def genes(genome, head=False):

    if not head:
        df = genes_df(genome)
    else:
        df = genes_df(
            genome,
            'select chrom, txStart, txEnd, exonStarts, exonEnds, name, name2, strand from refGene limit 500;'
        )

    return parse_genes(df)


def genes_df(
        genome,
        __query='select chrom, txStart, txEnd, exonStarts, exonEnds, name, name2, strand from refGene'
):

    df = ucsc(genome, __query)
    df.loc[:, "exonStarts"] = df["exonStarts"].str.decode("utf-8")
    df.loc[:, "exonEnds"] = df["exonEnds"].str.decode("utf-8")

    return df


def parse_genes(df):

    df.columns = "Chromosome Start End XS XE TranscriptID GeneID Strand".split(
    )

    df = df.astype({
        "Chromosome": "category",
        "TranscriptID": "category",
        "GeneID": "category",
        "Strand": "category",
        "Start": int,
        "End": int
    })

    exons = _exons(df)
    _df = (df.drop("XS XE".split(), axis=1).assign(Feature="transcript"))

    df = pd.concat([_df, exons], sort=False).sort_values(
        "Chromosome Start End".split()
    )["Chromosome Start End Strand Feature TranscriptID ExonNumber".split()]

    return pr.PyRanges(df)


def chromosome_sizes(genome):

    query = 'select chrom,size from chromInfo'

    df = ucsc(genome, query)
    df.columns = ["Chromosome", "Size"]
    s = pd.Series(data=df.Size.values, index=df.Chromosome.values)

    return s


def genomes():

    host = "genome-mysql.cse.ucsc.edu"
    user = "genome"

    conn = MySQLdb.Connection(host=host, user=user)

    df = pd.read_sql("show databases", conn)
    df.columns = ["Genome"]

    return df



import pandas as pd

from ftplib import FTP

from natsort import natsorted

from pyranges.db.methods import get_ftp_file


def genomes():

    hr = natsorted(_releases("human"))
    mr = natsorted(_releases("mouse"))

    hg = ["Human"] * len(hr)
    mg = ["Mouse"] * len(mr)

    v = pd.Series(hr + mr).str.extract("release_(\w+)")
    g = pd.Series(hg + mg)

    df = pd.concat([g, v], axis=1, sort=False)
    df.columns = "Genome Version".split()

    return df


def _releases(species="human"):

    assert species in "human mouse".split()

    ftp = FTP("ftp.ebi.ac.uk")
    ftp.login()
    ftp.cwd('pub/databases/gencode/Gencode_' + species)
    dir_listing = ftp.nlst()

    return [l for l in dir_listing if "release" in l]


def genes(species, release="latest", path=None, head=False):

    assert species in "human mouse".split()

    releases = []
    for r in _releases(species):
        r = r.split("_")[1]
        if r.replace("M", "").isdecimal():
            releases.append(r)

    if not release == "latest":
        assert str(
            release) in releases, str(release) + " not in " + str(releases)
    else:
        release = max(releases, key=lambda n: int(n.replace("M", "")))

    # ftp = FTP("ftp.ebi.ac.uk")
    # ftp.login()
    ftp_host = "ftp.ebi.ac.uk"

    binary_loc = 'pub/databases/gencode/Gencode_' + species + "/release_" + release + "/gencode.v{}.annotation.gtf.gz".format(
        release)

    return get_ftp_file(ftp_host, binary_loc, path, test=head)



from pyranges.subset import (get_string, get_tuple, get_slice, get_booldict)
from pyranges.methods.drop import _drop

from pyranges import PyRanges


def _getitem(self, val):

    if isinstance(val, list):
        dfs = _drop(self, keep=val).dfs
    elif isinstance(val, str):
        dfs = get_string(self, val)
    elif isinstance(val, tuple):
        dfs = get_tuple(self, val)
    elif isinstance(val, slice):
        dfs = get_slice(self, val)
    elif isinstance(val, dict):
        dfs = get_booldict(self, val)
    else:
        raise Exception("Not valid subsetter: {}".format(str(val)))

    if not dfs is None:
        return PyRanges(dfs)
    else:
        return PyRanges({})



import pytest

from hypothesis import given, settings, HealthCheck
from hypothesis import reproduce_failure  # pylint: disable=unused-import

from itertools import product
import tempfile
import subprocess  # nosec
from io import StringIO

import pandas as pd

from tests.helpers import assert_df_equal
from tests.hypothesis_helper import dfs_min2, dfs_min

from os import environ

if environ.get("TRAVIS"):
    max_examples = 100
    deadline = None
else:
    max_examples = 100
    deadline = None

strandedness = [False, "same", "opposite"]
no_opposite = [False, "same"]


def run_bedtools(command,
                 gr,
                 gr2,
                 strandedness,
                 nearest_overlap=False,
                 nearest_how=None):

    bedtools_strand = {False: "", "same": "-s", "opposite": "-S"}[strandedness]
    bedtools_overlap = {True: "", False: "-io"}[nearest_overlap]
    bedtools_how = {
        "upstream": "-id",
        "downstream": "-iu",
        None: ""
    }[nearest_how] + " -D a"
    # print("bedtools how:", bedtools_how)

    with tempfile.TemporaryDirectory() as temp_dir:
        f1 = "{}/f1.bed".format(temp_dir)
        f2 = "{}/f2.bed".format(temp_dir)
        gr.df.to_csv(f1, sep="\t", header=False, index=False)
        gr2.df.to_csv(f2, sep="\t", header=False, index=False)

        cmd = command.format(
            f1=f1,
            f2=f2,
            strand=bedtools_strand,
            overlap=bedtools_overlap,
            bedtools_how=bedtools_how)
        print("cmd " * 5)
        print(cmd)
        # ignoring the below line in bandit as only strings created by
        # the test suite is run here; no user input ever sought
        result = subprocess.check_output(  # nosec
            cmd, shell=True, executable="/bin/bash").decode()  #nosec

    return result


def read_bedtools_result_set_op(bedtools_result, strandedness):

    if strandedness:
        usecols = [0, 1, 2, 5]
        names = "Chromosome Start End Strand".split()
    else:
        usecols = [0, 1, 2]
        names = "Chromosome Start End".split()

    return pd.read_csv(
        StringIO(bedtools_result),
        header=None,
        usecols=usecols,
        names=names,
        sep="\t")


def compare_results(bedtools_df, result):

    if not bedtools_df.empty:
        assert_df_equal(result.df, bedtools_df)
    else:
        assert bedtools_df.empty == result.df.empty


def compare_results_nearest(bedtools_df, result):

    if not bedtools_df.empty:
        bedtools_df = bedtools_df[bedtools_df.Distance != -1]

    result = result.df

    if not len(result) == 0:
        result_df = result["Chromosome Start End Strand Distance".split()]
        assert_df_equal(result_df, bedtools_df)
    else:
        assert bedtools_df.empty


@pytest.mark.bedtools
@pytest.mark.parametrize("strandedness", no_opposite)
@settings(
    max_examples=max_examples,
    deadline=deadline,
    suppress_health_check=HealthCheck.all())
@given(gr=dfs_min(), gr2=dfs_min())  # pylint: disable=no-value-for-parameter
def test_set_intersect(gr, gr2, strandedness):

    set_intersect_command = "bedtools intersect {strand} -a <(sort -k1,1 -k2,2n {f1} | bedtools merge {strand} -c 4,5,6 -o first -i -) -b <(sort -k1,1 -k2,2n {f2} | bedtools merge {strand} -c 4,5,6 -o first -i -)"
    bedtools_result = run_bedtools(set_intersect_command, gr, gr2,
                                   strandedness)

    bedtools_df = read_bedtools_result_set_op(bedtools_result, strandedness)

    result = gr.set_intersect(gr2, strandedness=strandedness)

    compare_results(bedtools_df, result)


@pytest.mark.bedtools
@pytest.mark.parametrize("strandedness", no_opposite)
@settings(
    max_examples=max_examples,
    deadline=deadline,
    suppress_health_check=HealthCheck.all())
@given(gr=dfs_min(), gr2=dfs_min())  # pylint: disable=no-value-for-parameter
def test_set_union(gr, gr2, strandedness):

    set_union_command = "cat {f1} {f2} | bedtools sort | bedtools merge {strand} -c 4,5,6 -o first -i -"  # set_union_command = "bedtools merge {strand} -c 4,5,6 -o first -i {f1}"
    bedtools_result = run_bedtools(set_union_command, gr, gr2, strandedness)

    bedtools_df = read_bedtools_result_set_op(bedtools_result, strandedness)

    result = gr.set_union(gr2, strandedness=strandedness)

    compare_results(bedtools_df, result)


@pytest.mark.bedtools
@pytest.mark.parametrize("strandedness", strandedness)
@settings(
    max_examples=max_examples,
    deadline=deadline,
    suppress_health_check=HealthCheck.all())
@given(gr=dfs_min(), gr2=dfs_min())  # pylint: disable=no-value-for-parameter
def test_overlap(gr, gr2, strandedness):

    overlap_command = "bedtools intersect -wa {strand} -a {f1} -b {f2}"

    bedtools_result = run_bedtools(overlap_command, gr, gr2, strandedness)

    bedtools_df = pd.read_csv(
        StringIO(bedtools_result),
        header=None,
        names="Chromosome Start End Name Score Strand".split(),
        sep="\t")

    result = gr.overlap(gr2, strandedness=strandedness)

    compare_results(bedtools_df, result)


@pytest.mark.bedtools
@pytest.mark.parametrize("strandedness", strandedness)
@settings(
    max_examples=max_examples,
    deadline=deadline,
    suppress_health_check=HealthCheck.all())
@given(gr=dfs_min(), gr2=dfs_min())  # pylint: disable=no-value-for-parameter
# @reproduce_failure('4.5.7', b'AXicY2RAA4zoAgAAVQAD')
def test_intersect(gr, gr2, strandedness):

    intersect_command = "bedtools intersect {strand} -a {f1} -b {f2}"

    bedtools_result = run_bedtools(intersect_command, gr, gr2, strandedness)

    bedtools_df = pd.read_csv(
        StringIO(bedtools_result),
        header=None,
        names="Chromosome Start End Name Score Strand".split(),
        sep="\t")

    # from pydbg import dbg
    # dbg(gr)
    # dbg(gr2)

    result = gr.intersect(gr2, strandedness=strandedness)

    compare_results(bedtools_df, result)


# @pytest.mark.bedtools
# @pytest.mark.parametrize("strandedness", strandedness)
# @settings(
#     max_examples=max_examples,
#     deadline=deadline,
#     suppress_health_check=HealthCheck.all())
# @given(gr=dfs_min(), gr2=dfs_min())  # pylint: disable=no-value-for-parameter
# @reproduce_failure('4.15.0', b'AXicY2RgYGAEIzgAsRkZUfkMDAAA2AAI')
# def test_no_intersect(gr, gr2, strandedness):

#     intersect_command = "bedtools intersect -v {strand} -a {f1} -b {f2}"

#     bedtools_result = run_bedtools(intersect_command, gr, gr2, strandedness)

#     bedtools_df = pd.read_csv(
#         StringIO(bedtools_result),
#         header=None,
#         names="Chromosome Start End Name Score Strand".split(),
#         sep="\t")

#     # bedtools bug: https://github.com/arq5x/bedtools2/issues/719
#     result = gr.no_overlap(gr2, strandedness=strandedness)

#     from pydbg import dbg
#     dbg(result)
#     dbg(bedtools_df)

#     # result2 = gr.intersect(gr2, strandedness)

#     compare_results(bedtools_df, result)


@pytest.mark.bedtools
@pytest.mark.parametrize("strandedness", ["same", "opposite", False])  #
@settings(
    max_examples=max_examples,
    deadline=deadline,
    suppress_health_check=HealthCheck.all())
@given(gr=dfs_min(), gr2=dfs_min())  # pylint: disable=no-value-for-parameter
# @reproduce_failure('4.5.7', b'AXicLYaJCQAACIS0/YfuuQRRAbVG94Dk5LHSBgJ3ABU=')
def test_subtraction(gr, gr2, strandedness):

    subtract_command = "bedtools subtract {strand} -a {f1} -b {f2}"

    bedtools_result = run_bedtools(subtract_command, gr, gr2, strandedness)

    bedtools_df = pd.read_csv(
        StringIO(bedtools_result),
        header=None,
        names="Chromosome Start End Name Score Strand".split(),
        sep="\t")

    result = gr.subtract(gr2, strandedness=strandedness)

    compare_results(bedtools_df, result)


nearest_hows = [None, "upstream", "downstream"]
overlaps = [True, False]


@pytest.mark.bedtools
@pytest.mark.parametrize("nearest_how,overlap,strandedness",
                         product(nearest_hows, overlaps, strandedness))
@settings(
    max_examples=max_examples,
    deadline=deadline,
    suppress_health_check=HealthCheck.all())
@given(gr=dfs_min(), gr2=dfs_min())  # pylint: disable=no-value-for-parameter
# @reproduce_failure('4.15.0', b'AXicY2RgYGAEIygBp5FJEAONzwAAAZYACw==')
def test_nearest(gr, gr2, nearest_how, overlap, strandedness):

    nearest_command = "bedtools closest {bedtools_how} {strand} {overlap} -t first -d -a <(sort -k1,1 -k2,2n {f1}) -b <(sort -k1,1 -k2,2n {f2})"

    bedtools_result = run_bedtools(nearest_command, gr, gr2, strandedness,
                                   overlap, nearest_how)

    bedtools_df = pd.read_csv(
        StringIO(bedtools_result),
        header=None,
        names="Chromosome Start End Strand Chromosome2 Distance".split(),
        usecols=[0, 1, 2, 5, 6, 12],
        sep="\t")

    bedtools_df.Distance = bedtools_df.Distance.abs()

    bedtools_df = bedtools_df[bedtools_df.Chromosome2 != "."]
    bedtools_df = bedtools_df.drop("Chromosome2", 1)

    result = gr.nearest(
        gr2, strandedness=strandedness, overlap=overlap, how=nearest_how)

    print("bedtools " * 5)
    print(bedtools_df)
    print("result " * 5)
    print(result)

    compare_results_nearest(bedtools_df, result)


@pytest.mark.bedtools
@pytest.mark.parametrize("strandedness", no_opposite)
@settings(
    max_examples=max_examples,
    deadline=deadline,
    suppress_health_check=HealthCheck.all())
@given(gr=dfs_min(), gr2=dfs_min())  # pylint: disable=no-value-for-parameter
def test_jaccard(gr, gr2, strandedness):

    # jaccard_command = "bedtools jaccard {strand}  -a <(sort -k1,1 -k2,2n {f1}) -b <(sort -k1,1 -k2,2n {f2})"

    # bedtools_result = run_bedtools(jaccard_command, gr, gr2, strandedness)
    # print(bedtools_result)

    # bedtools_jaccard = float(bedtools_result.split("\n")[1].split()[2])
    # print(bedtools_jaccard)

    # https://github.com/arq5x/bedtools2/issues/645
    # will make tests proper when bedtools is fixed
    result = gr.stats.jaccard(gr2, strandedness=strandedness)

    # assert abs(result - bedtools_jaccard) < 0.001

    assert 0 <= result <= 1


@pytest.mark.bedtools
@pytest.mark.parametrize("strandedness", strandedness)
@settings(
    max_examples=max_examples,
    deadline=deadline,
    suppress_health_check=HealthCheck.all())
@given(gr=dfs_min(), gr2=dfs_min())  # pylint: disable=no-value-for-parameter
def test_join(gr, gr2, strandedness):

    join_command = "bedtools intersect {strand} -wo -a {f1} -b {f2}"

    bedtools_result = run_bedtools(join_command, gr, gr2, strandedness)

    bedtools_df = pd.read_csv(
        StringIO(bedtools_result),
        header=None,
        sep="\t",
        names=
        "Chromosome Start End Name Score Strand Chromosome_b Start_b End_b Name_b Score_b Strand_b Overlap"
        .split(),
        dtype={
            "Chromosome": "category",
            "Strand": "category"
        }).drop(
            "Chromosome_b Overlap".split(), axis=1)

    result = gr.join(gr2, strandedness=strandedness)

    if result.df.empty:
        assert bedtools_df.empty
    else:
        assert_df_equal(result.df, bedtools_df)


@pytest.mark.bedtools
@settings(
    max_examples=max_examples,
    deadline=deadline,
    suppress_health_check=HealthCheck.all())
@given(gr=dfs_min2(), gr2=dfs_min2())  # pylint: disable=no-value-for-parameter
# @reproduce_failure('4.5.7', b'AXicY2RgYGAEQihghLDBJCMjXJiBAQABIwAM')
# @reproduce_failure('4.5.7', b'AXicY2RgYGAEIwhgBDNBIkDMxAAHAAEIAAs=')
# @reproduce_failure('4.5.7', b'AXicY2RgYGAEIxhghIoAMRNCFAAA8AAK')
# @reproduce_failure('4.5.7', b'AXicY2RgYGAEIxhgRBNhZAAAALIACA==')
# above were working
# @reproduce_failure('4.5.7', b'AXicY2RgYGAEIiBmArPAgBFGMCLEgAwAAXUADQ==')
# @reproduce_failure('4.5.7', b'AXicY2RgYGAEIiBmYoACRpAIWBQhAgABMAAM')
# @reproduce_failure('4.5.7', b'AXicY2RgYGAEIwhgBDMRIkCKGUQDAAD7AAw=')
# above were working
# @reproduce_failure('4.5.7', b'AXicY2RgYGAEIiBmYgADRjCCEDA2BAAAAWcADA==')
# @reproduce_failure('4.5.7', b'AXicY2RgYGCEIQYmBihghAszILEAAU4ADA==')
#
# @reproduce_failure('4.5.7', b'AXicY2RgYGAEIwhgBDNBIkDMhJBgAhEAAVYADg==')
#
# @reproduce_failure('4.5.7', b'AXicY2RgYGCEITBghBEgMSBmgvCYoNIMAAGdAA8=')
def test_reldist(gr, gr2):

    reldist_command = "bedtools reldist -a <(sort -k1,1 -k2,2n {f1}) -b <(sort -k1,1 -k2,2n {f2})"

    bedtools_result = run_bedtools(reldist_command, gr, gr2, False)
    bedtools_result = pd.read_csv(StringIO(bedtools_result), sep="\t")
    print(bedtools_result)

    result = gr.stats.relative_distance(gr2)
    print(result)

    # bug in bedtools, therefore not testing this properly
    # https://github.com/arq5x/bedtools2/issues/711

    assert 1


new_pos = ["union"]  # ["intersection", "union"]


@pytest.mark.parametrize("strandedness,new_pos", product(
    strandedness, new_pos))
@settings(
    max_examples=max_examples,
    deadline=deadline,
    suppress_health_check=HealthCheck.all())
@given(gr=dfs_min(), gr2=dfs_min())  # pylint: disable=no-value-for-parameter
def test_join_new_pos(gr, gr2, strandedness, new_pos):

    result = gr.join(gr2, strandedness=strandedness, new_pos=new_pos)

    import numpy as np
    result2 = gr.join(gr2, strandedness=strandedness)

    if result.df.empty:
        assert result2.df.empty
    else:
        if new_pos == "union":
            new_starts = np.minimum(result2.Start, result2.Start_b)
            new_ends = np.maximum(result2.End, result2.End_b)
        else:
            new_starts = np.maximum(result2.Start, result2.Start_b)
            new_ends = np.minimum(result2.End, result2.End_b)
        assert list(result.Start.values) == list(new_starts)
        assert list(result.End.values) == list(new_ends)



import pyranges.raymock as ray

import numpy as np
import pandas as pd
from ncls import NCLS


@ray.remote
def _intersection(scdf, ocdf, kwargs):

    how = kwargs["how"]

    if ocdf.empty or scdf.empty:
        return None

    assert how in "containment first".split() + [False, None]
    starts = scdf.Start.values
    ends = scdf.End.values
    indexes = scdf.index.values

    in_dtype = ocdf.Start.dtype

    oncls = NCLS(ocdf.Start.values, ocdf.End.values, ocdf.index.values)

    if not how or how is None:
        _self_indexes, _other_indexes = oncls.all_overlaps_both(
            starts, ends, indexes)
    elif how == "containment":
        _self_indexes, _other_indexes = oncls.all_containments_both(
            starts, ends, indexes)
    elif how == "first":
        _self_indexes, _other_indexes = oncls.first_overlap_both(
            starts, ends, indexes)

    _self_indexes = _self_indexes
    _other_indexes = _other_indexes

    scdf, ocdf = scdf.reindex(_self_indexes), ocdf.reindex(_other_indexes)

    new_starts = pd.Series(
        np.where(scdf.Start.values > ocdf.Start.values, scdf.Start,
                 ocdf.Start),
        index=scdf.index,
        dtype=in_dtype)

    new_ends = pd.Series(
        np.where(scdf.End.values < ocdf.End.values, scdf.End, ocdf.End),
        index=scdf.index,
        dtype=in_dtype)

    pd.options.mode.chained_assignment = None  # default='warn'
    scdf.loc[:, "Start"] = new_starts
    scdf.loc[:, "End"] = new_ends
    pd.options.mode.chained_assignment = 'warn'

    if not scdf.empty:
        return scdf
    else:
        return None


@ray.remote
def _overlap(scdf, ocdf, kwargs):

    invert = kwargs["invert"]

    if scdf.empty or ocdf.empty:
        return None

    how = kwargs["how"]

    assert how in "containment first".split() + [False, None]
    starts = scdf.Start.values
    ends = scdf.End.values
    indexes = scdf.index.values

    it = NCLS(ocdf.Start.values, ocdf.End.values, ocdf.index.values)

    if not how:
        _indexes = it.all_overlaps_self(starts, ends, indexes)
    elif how == "containment":
        _indexes = it.has_containment(starts, ends, indexes)
    else:
        _indexes = it.has_overlaps(starts, ends, indexes)

    # dbg(_indexes)

    if invert:
        _indexes = scdf.index.difference(_indexes)
    # dbg(_indexes)
    # dbg(scdf)

    return scdf.reindex(_indexes)


# @ray.remote
# def _first_df(scdf, ocdf, kwargs):

#     if scdf.empty or ocdf.empty:
#         return None

#     how = kwargs["how"]

#     assert how in "containment first".split() + [False, None]
#     starts = scdf.Start.values
#     ends = scdf.End.values
#     indexes = scdf.index.values

#     it = NCLS(ocdf.Start.values, ocdf.End.values, ocdf.index.values)

#     if not how:
#         _indexes = it.has_overlaps(starts, ends, indexes)
#     elif how == "containment":
#         _indexes = it.has_containment(starts, ends, indexes)

#     return scdf.reindex(_indexes)



import re
from collections.abc import Iterable


def _drop(self, drop=None, keep=None, drop_strand=False):
    columns = self.columns
    if drop_strand:
        always_keep = "Chromosome Start End".split()
    else:
        always_keep = "Chromosome Start End Strand".split()

    _to_drop = []

    if not drop and not keep:
        _to_drop = set(columns)
    elif isinstance(drop, str):
        r = re.compile(drop)
        _to_drop = [c for c in columns if not r.search(c) == None]
    elif isinstance(drop, Iterable) or isinstance(drop, list):
        _to_drop = drop
    elif isinstance(keep, str):
        r = re.compile(keep)
        _to_drop = [c for c in columns if r.search(c) is None]
    elif isinstance(keep, Iterable) or isinstance(keep, list):
        _to_drop = set(columns) - set(keep) - set(always_keep)
    else:
        raise Exception("Not valid subsetters!")

    if not drop_strand:
        assert not set(always_keep).intersection(_to_drop), \
            "Can never drop Chromosome, Start or End. Cannot drop Strand unless drop_strand is True"
    _to_drop = set(_to_drop) - set(always_keep)

    return self.apply(lambda df: df.drop(_to_drop, axis=1))



import pytest


def test_getsetattr(chip_10):

    chip_10.whatevz = list(range(len(chip_10)))

    assert (chip_10.whatevz == list(range(len(chip_10)))).all()


def test_getsetattr_fails(chip_10):

    with pytest.raises(Exception):
        chip_10.whatevz = range(len(chip_10) + 1)


def test_getsetattr_with_str(chip_10):

    chip_10.whatevz = "whatevz"

    assert (chip_10.whatevz == list(["whatevz"] * len(chip_10))).all()



import os
import pandas as pd

import MySQLdb

from pyranges.db.methods import get_ftp_file


def _find_correct_db_version(conn, genome, version):
    """ensembl dbs have names like danio_rerio_75_11 where the user has no way of knowing the last
magica number. This function connects to the correct database and version number (danio_rerio_75)."""

    query = 'show databases like "{}_core_%"'.format(genome)

    df = pd.read_sql(query, conn)

    if df.empty:
        raise Exception(
            "Genome {} not found. List all possible genomes and versions with pyranges.db.ensembl.genomes()"
            .format(genome))

    df.columns = ["db"]

    df = df[(df.db.str.contains(genome)) & (df.db.str.contains("_core_"))]

    versions = df.db.str.extract(genome +
                                 "_core_([0-9]+)_*").astype(int).squeeze()

    if version == "latest":
        idx = versions.idxmax()
        db = df.loc[idx].iloc[0]
    else:
        idx = versions[version == versions].index

        assert len(
            idx), "No version {} in list of ensembl versions: {}".format(
                version, versions)
        assert len(
            idx) == 1, "Multiple rows matching version number {} in {}".format(
                version, df)

        db = df.loc[idx].iloc[0, 0]

    cursor = conn.cursor()
    cursor.execute("use {}".format(db))

    return conn


def ensembl(genome, query, version="latest"):

    host = "ensembldb.ensembl.org"
    user = "anonymous"
    conn = MySQLdb.Connection(host=host, user=user)

    conn = _find_correct_db_version(conn, genome, version)

    df = pd.read_sql(query, conn)

    conn.close()

    return df


def chromosome_sizes(genome, version="latest"):

    query = 'select name,length from seq_region'

    df = ensembl(genome, query, version)

    return pd.Series(data=df.length.values, index=df.name.values)


def genomes():

    host = "ensembldb.ensembl.org"
    user = "anonymous"
    conn = MySQLdb.Connection(host=host, user=user)

    query = "show databases like '%_core_%'"
    df = pd.read_sql(query, conn)
    df.columns = ["db"]

    df = df.db.str.extract("(\w+)_core_(\d+)_*.")
    df.columns = ["Genome", "Version"]

    return df


# ftp://ftp.ensembl.org/pub/release-95/gtf/gopherus_agassizii

from ftplib import FTP


def genes(genome, release="latest", path=None, head=False):

    ftp = FTP("ftp.ensembl.org")
    ftp.login()

    if release == "latest":
        release = max(
            [r.split("-")[1] for r in ftp.nlst("pub/") if "release-" in r],
            key=int)

    _dir = "pub/release-{}/gtf/{}/".format(release, genome)
    dir_listing = pd.Series(ftp.nlst(_dir))

    matches = (dir_listing.str.lower().apply(os.path.basename).str.match(
        "{}.*{}.gtf.gz".format(genome, release)))

    assert matches.sum() == 1, "More than one file matching: {}".format(
        dir_listing[matches])

    if not len(dir_listing):
        raise Exception(
            "No files found for genome {}. Use pyranges.db.ensembl.genomes() to list all available genomes."
            .format(genome))

    binary_loc = dir_listing[matches].iloc[0]

    host = "ftp.ensembl.org"
    return get_ftp_file(host, binary_loc, path, test=head)



import pytest

from hypothesis import given, settings, HealthCheck
from hypothesis import reproduce_failure  # pylint: disable=unused-import

from itertools import product
import tempfile
import subprocess  # nosec
from io import StringIO

import pandas as pd

from tests.helpers import assert_df_equal
from tests.hypothesis_helper import dfs_min2, dfs_min, dfs_no_min

from os import environ

from itertools import product

if environ.get("TRAVIS"):
    max_examples = 10
    deadline = None
else:
    max_examples = 100
    deadline = None

strandedness = [False, "same", "opposite"]

binary_methods = [
    "set_union", "set_intersect", "overlap", "nearest", "intersect",
    "subtract", "join"
]

unary_methods = ["merge", "sort"]

method_chain = product(binary_methods, binary_methods)
# cannot start with an operation that makes pyrange unstranded and then try a stranded op
strandedness_chain = list(product(["same", "opposite"], strandedness)) + list(
    product(strandedness, [None]))


@pytest.mark.bedtools
@pytest.mark.parametrize("strandedness_chain,method_chain",
                         product(strandedness_chain, method_chain))
@settings(
    max_examples=max_examples,
    deadline=deadline,
    suppress_health_check=HealthCheck.all())
@given(gr=dfs_no_min(), gr2=dfs_no_min(), gr3=dfs_no_min())  # pylint: disable=no-value-for-parameter
def test_three_in_a_row(gr, gr2, gr3, strandedness_chain, method_chain):

    s1, s2 = strandedness_chain
    f1, f2 = method_chain

    # print(s1, s2)
    # print(f1, f2)

    m1 = getattr(gr, f1)
    gr2 = m1(gr2, strandedness=s1)
    m2 = getattr(gr2, f2)
    gr3 = m2(gr3, strandedness=s2)
